from .centerline import centerline
import numpy as np
from shapely.geometry import LineString, Point
import geopandas as gpd
import pandas as pd


class XSGen:
    """ The XSGen class generates a cross-section on a stream segment given a location,
        width number of points. It fits a tension spline to stream segment, and
        calculated cross-section perpendicular from sline.
    """

    def __init__(self, point, cl_geom, ny, width) -> None:
        
        self.cl_geom = cl_geom
        self.point = point
        self.tension = 0.5
        self.width = width
        self.cl_length = self.cl_geom.geometry[0].length
        if ny % 2 == 0:
            ny += 1
        self.ny = ny
        if self.cl_length > 20.0:
            self.nx = int(self.cl_length / 10)
        else:
            self.nx = int(self.cl_length / 1)
        self.cl = centerline(cl_geom, self.nx, self.tension)
        self.x = np.zeros(self.ny, dtype=np.double)
        self.y = np.zeros(self.ny, dtype=np.double)
        self._buildxs()

    def _get_perp_index(self, clx, cly):
        mind = 1e6
        id = -1
        for index, p in enumerate(zip(clx, cly)):
            dist = np.sqrt(
                np.power(self.point.geometry[0].x - p[0], 2)
                + np.power(self.point.geometry[0].y - p[1], 2)
            )
            if dist < mind:
                mind = dist
                id = index

        return id

    def _buildxs(self):
        clx, cly = self.cl.getinterppts()
        delt = np.double(self.width / (self.ny - 1))
        nm = int((self.ny + 1) / 2)
        index = self._get_perp_index(clx, cly)

        for id, j in enumerate(range(0, self.ny)):
            self.x[id] = clx[index] + delt * (nm - j - 1) * np.sin(
                self.cl.getphiinterp(index)
            )
            self.y[id] = cly[index] - delt * (nm - j - 1) * np.cos(
                self.cl.getphiinterp(index)
            )

    def get_xs(self):
        points = gpd.GeoSeries(map(Point, zip(self.x, self.y)))
        ls = LineString((points.to_list()))
        d = {0: {"name": "cross-section", "geometry": ls}}
        df = pd.DataFrame.from_dict(d, orient="index")
        gdf = gpd.GeoDataFrame(df, geometry=df.geometry, crs=self.point.crs)
        return gdf

    def get_xs_points(self):
        return self.x, self.y

    def get_strm_seg_spline(self):
        x, y = self.cl.getinterppts()
        points = gpd.GeoSeries(map(Point, zip(x, y)))
        ls = LineString((points.to_list()))
        d = {0: {"name": "strm_seg_spline", "geometry": ls}}
        df = pd.DataFrame.from_dict(d, orient="index")
        gdf = gpd.GeoDataFrame(df, geometry=df.geometry)
        return gdf



'''Console script for nldi_el_serv.'''
import sys
import click
# import numpy as np
from nldi_el_serv.nldi_el_serv import getXSAtPoint, getXSAtEndPts
# import geopandas as gpd
# import pandas as pd
# import json

resdict = {'1m': 1, '3m': 3, '5m': 5, '10m': 10, '30m': 30, '60m': 60}


class NLDI_El_Serv:
    def __init__(self):
        self.out_crs = 'epsg:4326'

    def setoutCRS(self, out_crs='epsg:4326'):
        self.out_crs = out_crs

    def outCRS(self):
        return self.out_crs

    def __repr__(self):
        return f'NLDI_El_Serv {self.out_crs}'


pass_nldi_el_serv = click.make_pass_decorator(NLDI_El_Serv)


@click.group()
@click.option('--outcrs',
              default='epsg:4326',
              help='Projection CRS to return cross-section geometry: default is epsg:4326')
@click.version_option('0.1')
@click.pass_context
def main(ctx, outcrs):
    '''NLDI_El_Serv is a command line tool to for elevation-based services to the NLDI'''

    ctx.obj = NLDI_El_Serv()
    ctx.obj.setoutCRS(outcrs)
    return 0

# XS command at point with NHD


@main.command()
@click.option(
                '-f', '--file',
                default=None,
                type=click.File('w'),
                help='enter path and filenmae for json ouput'
             )
@click.option(
                '-ll', '--lonlat',
                required=True,
                type=tuple((float, float)),
                help='format lon,lat (x,y) as floats for example: -103.8011 40.2684'
             )
@click.option(
                '-n', '--numpoints',
                default=101,
                type=int,
                help='number of points in cross-section'
             )
@click.option(
                '-w', '--width',
                default=1000.0,
                type=float,
                help='width of cross-section')
@click.option(
                '-r', '--resolution',
                type=click.Choice(['1m', '3m', '5m', '10m', '30m', '60m'], case_sensitive=False),
                default='10m',
                help='Resolution of DEM used.  Note: 3DEP provides server side interpolatin given best available data'
             )
@click.option(
                '-v', '--verbose',
                default=False,
                type=bool,
                help='verbose ouput'
             )
@pass_nldi_el_serv
def XSAtPoint(nldi_el_serv, lonlat, numpoints, width, resolution, file, verbose):
    x = lonlat[0]
    y = lonlat[1]
    nl = '\n'
    if verbose:
        print(
                f'input={lonlat}, lat={x}, lon={y}, {nl} \
                npts={numpoints}, width={width}, resolution={resolution}, {nl} \
                crs={nldi_el_serv.outCRS()}, {nl} \
                file={file}, {nl} \
                out_epsg={nldi_el_serv.outCRS()}'
            )
    # print(tuple(latlon))
    xs = getXSAtPoint(
                        point=tuple((x, y)),
                        numpoints=numpoints,
                        width=width,
                        file=file,
                        res=resdict.get(resolution)
                    )
    if file is None:
        print(xs.to_json())
    return 0

# XS command at user defined endpoints


@main.command()
@click.option('-f', '--file',
              default=None,
              type=click.File('w'),
              help='Output json file')
@click.option('-s', '--startpt',
              required=True,
              type=tuple((float, float)),
              help='format x y pair as floats for example: -103.801134 40.267335')
@click.option('-e', '--endpt',
              required=True,
              type=tuple((float, float)),
              help='format x y pair as floats for example: -103.800787 40.272798 ')
@click.option('-c', '--crs',
              required=True,
              type=str,
              help='spatial reference of input data',
              default='epsg:4326')
@click.option('-n', '--numpoints',
              default=100,
              type=int,
              help='number of points in cross-section')
@click.option(
                '-r', '--resolution',
                type=click.Choice(['1m', '3m', '5m', '10m', '30m', '60m'], case_sensitive=False),
                default='10m',
                help='Resolution of DEM used.  Note: 3DEP provides server side interpolatin given best available data'
             )
@click.option('-v', '--verbose',
              default=False,
              type=bool,
              help='verbose ouput')
@pass_nldi_el_serv
def XSAtEndPts(nldi_el_serv, startpt, endpt, crs, numpoints, resolution, file, verbose):
    x1 = startpt[0]
    y1 = startpt[1]
    x2 = endpt[0]
    y2 = endpt[1]
    nl = '\n'
    if verbose:
        print(
            f'input:  {nl}, \
            start: {startpt}, {nl}, \
            end: {endpt}, {nl}, \
            x1:{x1}, y1:{y1}, {nl}, \
            x2:{x2}, y2:{y2}, {nl}, \
            npts={numpoints}, {nl}, \
            resolution={resolution}, \
            input_crs={crs}, {nl}, \
            output_crs={nldi_el_serv.outCRS()}  {nl}, \
            file={file}, {nl}, \
            verbose: {verbose} '
            )
    path = []
    path.append(startpt)
    path.append(endpt)
    # print(type(path))
    xs = getXSAtEndPts(
                        path=path,
                        numpts=numpoints,
                        res=resdict.get(resolution),
                        crs=crs,
                        file=file
                      )
    if file is None:
        print(xs.to_json())
    return 0


if __name__ == '__main__':
    sys.exit(main())  # pragma: no cover



# -*- coding: utf-8 -*-
"""
Created on Tue May  5 16:26:25 2015

@author: mweier
"""

import numpy as np
from numba import jit


@jit
def channelBuilder(wsDepth, rightSS, leftSS, widthBottom):
    """
    Builds trapziodal channel station/elevation array given depth,
    right side slope, left side slope, and bottom width
    """
    leftToe = wsDepth*1.25*leftSS
    rightToe = wsDepth*1.25*rightSS
    staElev = np.array([(0.0, wsDepth*1.25),
                        (leftToe, 0.0),
                        (leftToe + widthBottom, 0.0),
                        (leftToe+widthBottom+rightToe, wsDepth*1.25)])
    return staElev


def lineIntersection(line1, line2):
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        x = y = np.nan
#        print 'lines do not intersect'
        return x, y

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return x, y


@jit
def polygonArea(corners):
    area = 0.0
    for i in range(len(corners)):
        j = (i + 1) % len(corners)
        area += corners[i][0] * corners[j][1]
        area -= corners[j][0] * corners[i][1]
    area = abs(area) / 2.0
    return area


@jit
def channelPerimeter(corners):
    P = 0.0
    for i in range(len(corners)-1):
        P += np.sqrt((np.power((corners[i+1][0]-corners[i][0]), 2) +
                      np.power((corners[i+1][1]-corners[i][1]), 2)))
    return P


def flowEst(wsElev, n, slope, staElev, units):
    """
    Estimates uniform flow using the Manning equation for
    a user defined trapziodal channel or a manually defined channel using
    a station/elevation file
    """

    if units == "m":
        const = 1.0
    else:
        const = 1.49

    intersectList = []
    for i in range(0, len(staElev)):
        x, y = lineIntersection(
            (staElev[i-1], staElev[i]),
            ([staElev[0][0], wsElev], [staElev[-1][0], wsElev]))
        if x >= staElev[i-1][0] and x <= staElev[i][0] and abs(y - wsElev) < 0.01:
            #             print (x,y)
            intersectList.append((x, y))
        else:
            #             print ('line segments do not intersect')
            pass

    try:
        intersectArray = np.array(intersectList)
        intersectArray = intersectArray[intersectArray[:, 0].argsort()]
        # print 'more than two points intersect'
        staMinElev = staElev[np.where(
            staElev[:, 1] == min(staElev[:, 1]))][0][0]
        startPoint = intersectArray[np.where(
            intersectArray[:, 0] < staMinElev)][-1]
        endPoint = intersectArray[np.where(
            intersectArray[:, 0] > staMinElev)][0]
        intersectArray = np.vstack([startPoint, endPoint])
    except Exception as e:
        print(e)
        return 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0

    staMin = np.min(intersectArray[:, 0])
    staMax = np.max(intersectArray[:, 0])

    thalweig = staElev[np.where(staElev[:, 1] == np.min(staElev[:, 1]))]

    minElev = thalweig[:, 1][0]
    maxDepth = wsElev-minElev

    if len(intersectArray) < 2:
        return 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0

    staElevTrim = np.vstack([intersectArray[0], staElev, intersectArray[1]])
    # staElevTrim = staElevTrim[staElevTrim[:,0].argsort()]
    staElevTrim = staElevTrim[np.where(
        (staElevTrim[:, 0] >= staMin) & (staElevTrim[:, 0] <= staMax))]

    area = polygonArea(staElevTrim)
    R = area/channelPerimeter(staElevTrim)
    v = (const/n)*np.power(R, (2./3.0))*np.sqrt(slope)
    Q = v*area
    topWidth = staMax-staMin
    xGround = staElev[:, 0]
    yGround = staElev[:, 1]
    yGround0 = np.ones(len(xGround))*np.min(yGround)
    xWater = staElevTrim[:, 0]
    yWater = np.ones(len(xWater))*wsElev
    yWater0 = staElevTrim[:, 1]
    args = R, area, topWidth, Q, v, maxDepth, xGround, yGround, yGround0, xWater, yWater, yWater0
    return args
